@@ -5,9 +5,10 @@
 #include <time.h>
 #include <wlr/backend/interface.h>
 #include <wlr/backend/session.h>
+#include <wlr/types/wlr_buffer.h>
 #include <wlr/util/log.h>
+#include "backend/backend.h"
 #include "backend/multi.h"
-#include "util/signal.h"
 
 struct subbackend_state {
 	struct wlr_backend *backend;
@@ -58,24 +59,10 @@ static void multi_backend_destroy(struct wlr_backend *wlr_backend) {
 	}
 
 	// Destroy this backend only after removing all sub-backends
-	wlr_signal_emit_safe(&wlr_backend->events.destroy, backend);
+	wlr_backend_finish(wlr_backend);
 	free(backend);
 }
 
-static struct wlr_renderer *multi_backend_get_renderer(
-		struct wlr_backend *backend) {
-	struct wlr_multi_backend *multi = multi_backend_from_backend(backend);
-
-	struct subbackend_state *sub;
-	wl_list_for_each(sub, &multi->backends, link) {
-		struct wlr_renderer *rend = wlr_backend_get_renderer(sub->backend);
-		if (rend != NULL) {
-			return rend;
-		}
-	}
-	return NULL;
-}
-
 static struct wlr_session *multi_backend_get_session(
 		struct wlr_backend *_backend) {
 	struct wlr_multi_backend *backend = multi_backend_from_backend(_backend);
@@ -96,12 +83,48 @@ static clockid_t multi_backend_get_presentation_clock(
 	return CLOCK_MONOTONIC;
 }
 
-struct wlr_backend_impl backend_impl = {
+static int multi_backend_get_drm_fd(struct wlr_backend *backend) {
+	struct wlr_multi_backend *multi = multi_backend_from_backend(backend);
+
+	struct subbackend_state *sub;
+	wl_list_for_each(sub, &multi->backends, link) {
+		if (sub->backend->impl->get_drm_fd) {
+			return wlr_backend_get_drm_fd(sub->backend);
+		}
+	}
+
+	return -1;
+}
+
+static uint32_t multi_backend_get_buffer_caps(struct wlr_backend *backend) {
+	struct wlr_multi_backend *multi = multi_backend_from_backend(backend);
+
+	if (wl_list_empty(&multi->backends)) {
+		return 0;
+	}
+
+	uint32_t caps = WLR_BUFFER_CAP_DATA_PTR | WLR_BUFFER_CAP_DMABUF
+			| WLR_BUFFER_CAP_SHM;
+
+	struct subbackend_state *sub;
+	wl_list_for_each(sub, &multi->backends, link) {
+		uint32_t backend_caps = backend_get_buffer_caps(sub->backend);
+		if (backend_caps != 0) {
+			// only count backend capable of presenting a buffer
+			caps = caps & backend_caps;
+		}
+	}
+
+	return caps;
+}
+
+static const struct wlr_backend_impl backend_impl = {
 	.start = multi_backend_start,
 	.destroy = multi_backend_destroy,
-	.get_renderer = multi_backend_get_renderer,
 	.get_session = multi_backend_get_session,
 	.get_presentation_clock = multi_backend_get_presentation_clock,
+	.get_drm_fd = multi_backend_get_drm_fd,
+	.get_buffer_caps = multi_backend_get_buffer_caps,
 };
 
 static void handle_display_destroy(struct wl_listener *listener, void *data) {
@@ -137,13 +160,13 @@ bool wlr_backend_is_multi(struct wlr_backend *b) {
 static void new_input_reemit(struct wl_listener *listener, void *data) {
 	struct subbackend_state *state = wl_container_of(listener,
 			state, new_input);
-	wlr_signal_emit_safe(&state->container->events.new_input, data);
+	wl_signal_emit_mutable(&state->container->events.new_input, data);
 }
 
 static void new_output_reemit(struct wl_listener *listener, void *data) {
 	struct subbackend_state *state = wl_container_of(listener,
 			state, new_output);
-	wlr_signal_emit_safe(&state->container->events.new_output, data);
+	wl_signal_emit_mutable(&state->container->events.new_output, data);
 }
 
 static void handle_subbackend_destroy(struct wl_listener *listener,
@@ -165,6 +188,9 @@ static struct subbackend_state *multi_backend_get_subbackend(struct wlr_multi_ba
 
 bool wlr_multi_backend_add(struct wlr_backend *_multi,
 		struct wlr_backend *backend) {
+	assert(_multi && backend);
+	assert(_multi != backend);
+
 	struct wlr_multi_backend *multi = multi_backend_from_backend(_multi);
 
 	if (multi_backend_get_subbackend(multi, backend)) {
@@ -172,15 +198,6 @@ bool wlr_multi_backend_add(struct wlr_backend *_multi,
 		return true;
 	}
 
-	struct wlr_renderer *multi_renderer =
-		multi_backend_get_renderer(&multi->backend);
-	struct wlr_renderer *backend_renderer = wlr_backend_get_renderer(backend);
-	if (multi_renderer != NULL && backend_renderer != NULL && multi_renderer != backend_renderer) {
-		wlr_log(WLR_ERROR, "Could not add backend: multiple renderers at the "
-			"same time aren't supported");
-		return false;
-	}
-
 	struct subbackend_state *sub = calloc(1, sizeof(struct subbackend_state));
 	if (sub == NULL) {
 		wlr_log(WLR_ERROR, "Could not add backend: allocation failed");
@@ -200,7 +217,7 @@ bool wlr_multi_backend_add(struct wlr_backend *_multi,
 	wl_signal_add(&backend->events.new_output, &sub->new_output);
 	sub->new_output.notify = new_output_reemit;
 
-	wlr_signal_emit_safe(&multi->events.backend_add, backend);
+	wl_signal_emit_mutable(&multi->events.backend_add, backend);
 	return true;
 }
 
@@ -212,7 +229,7 @@ void wlr_multi_backend_remove(struct wlr_backend *_multi,
 		multi_backend_get_subbackend(multi, backend);
 
 	if (sub) {
-		wlr_signal_emit_safe(&multi->events.backend_remove, backend);
+		wl_signal_emit_mutable(&multi->events.backend_remove, backend);
 		subbackend_state_destroy(sub);
 	}
 }
