@@ -5,22 +5,21 @@
 #include <time.h>
 #include <wayland-server-core.h>
 #include <wlr/types/wlr_data_device.h>
-#include <wlr/types/wlr_input_device.h>
 #include <wlr/types/wlr_primary_selection.h>
 #include <wlr/types/wlr_seat.h>
+#include <wlr/types/wlr_compositor.h>
 #include <wlr/util/log.h>
 #include "types/wlr_seat.h"
 #include "util/global.h"
-#include "util/signal.h"
 
-#define SEAT_VERSION 7
+#define SEAT_VERSION 8
 
 static void seat_handle_get_pointer(struct wl_client *client,
 		struct wl_resource *seat_resource, uint32_t id) {
 	struct wlr_seat_client *seat_client =
 		wlr_seat_client_from_resource(seat_resource);
 	if (!(seat_client->seat->accumulated_capabilities & WL_SEAT_CAPABILITY_POINTER)) {
-		wl_resource_post_error(seat_resource, 0,
+		wl_resource_post_error(seat_resource, WL_SEAT_ERROR_MISSING_CAPABILITY,
 				"wl_seat.get_pointer called when no pointer capability has existed");
 		return;
 	}
@@ -34,7 +33,7 @@ static void seat_handle_get_keyboard(struct wl_client *client,
 	struct wlr_seat_client *seat_client =
 		wlr_seat_client_from_resource(seat_resource);
 	if (!(seat_client->seat->accumulated_capabilities & WL_SEAT_CAPABILITY_KEYBOARD)) {
-		wl_resource_post_error(seat_resource, 0,
+		wl_resource_post_error(seat_resource, WL_SEAT_ERROR_MISSING_CAPABILITY,
 				"wl_seat.get_keyboard called when no keyboard capability has existed");
 		return;
 	}
@@ -48,7 +47,7 @@ static void seat_handle_get_touch(struct wl_client *client,
 	struct wlr_seat_client *seat_client =
 		wlr_seat_client_from_resource(seat_resource);
 	if (!(seat_client->seat->accumulated_capabilities & WL_SEAT_CAPABILITY_TOUCH)) {
-		wl_resource_post_error(seat_resource, 0,
+		wl_resource_post_error(seat_resource, WL_SEAT_ERROR_MISSING_CAPABILITY,
 				"wl_seat.get_touch called when no touch capability has existed");
 		return;
 	}
@@ -67,7 +66,7 @@ static void seat_client_handle_resource_destroy(
 		return;
 	}
 
-	wlr_signal_emit_safe(&client->events.destroy, client);
+	wl_signal_emit_mutable(&client->events.destroy, client);
 
 	if (client == client->seat->pointer_state.focused_client) {
 		client->seat->pointer_state.focused_client = NULL;
@@ -76,6 +75,10 @@ static void seat_client_handle_resource_destroy(
 		client->seat->keyboard_state.focused_client = NULL;
 	}
 
+	if (client->seat->drag && client == client->seat->drag->seat_client) {
+		client->seat->drag->seat_client = NULL;
+	}
+
 	struct wl_resource *resource, *tmp;
 	wl_resource_for_each_safe(resource, tmp, &client->pointers) {
 		wl_resource_destroy(resource);
@@ -143,6 +146,20 @@ static void seat_handle_bind(struct wl_client *client, void *_wlr_seat,
 		wl_signal_init(&seat_client->events.destroy);
 
 		wl_list_insert(&wlr_seat->clients, &seat_client->link);
+
+		struct wlr_surface *pointer_focus =
+			wlr_seat->pointer_state.focused_surface;
+		if (pointer_focus != NULL &&
+				wl_resource_get_client(pointer_focus->resource) == client) {
+			wlr_seat->pointer_state.focused_client = seat_client;
+		}
+
+		struct wlr_surface *keyboard_focus =
+			wlr_seat->keyboard_state.focused_surface;
+		if (keyboard_focus != NULL &&
+				wl_resource_get_client(keyboard_focus->resource) == client) {
+			wlr_seat->keyboard_state.focused_client = seat_client;
+		}
 	}
 
 	wl_resource_set_implementation(wl_resource, &seat_impl,
@@ -159,7 +176,17 @@ void wlr_seat_destroy(struct wlr_seat *seat) {
 		return;
 	}
 
-	wlr_signal_emit_safe(&seat->events.destroy, seat);
+	wlr_seat_pointer_clear_focus(seat);
+	wlr_seat_keyboard_clear_focus(seat);
+
+	wlr_seat_set_keyboard(seat, NULL);
+
+	struct wlr_touch_point *point;
+	wl_list_for_each(point, &seat->touch_state.touch_points, link) {
+		wlr_seat_touch_point_clear_focus(seat, 0, point->touch_id);
+	}
+
+	wl_signal_emit_mutable(&seat->events.destroy, seat);
 
 	wl_list_remove(&seat->display_destroy.link);
 
@@ -183,7 +210,7 @@ void wlr_seat_destroy(struct wlr_seat *seat) {
 		}
 	}
 
-	wlr_global_destroy_safe(seat->global, seat->display);
+	wlr_global_destroy_safe(seat->global);
 	free(seat->pointer_state.default_grab);
 	free(seat->keyboard_state.default_grab);
 	free(seat->touch_state.default_grab);
