@@ -4,10 +4,9 @@
 #include <string.h>
 #include <time.h>
 #include <wayland-server-core.h>
-#include <wlr/types/wlr_input_device.h>
+#include <wlr/types/wlr_compositor.h>
 #include <wlr/util/log.h>
 #include "types/wlr_seat.h"
-#include "util/signal.h"
 
 static uint32_t default_touch_down(struct wlr_seat_touch_grab *grab,
 		uint32_t time, struct wlr_touch_point *point) {
@@ -33,16 +32,27 @@ static void default_touch_enter(struct wlr_seat_touch_grab *grab,
 	// not handled by default
 }
 
+static void default_touch_frame(struct wlr_seat_touch_grab *grab) {
+	wlr_seat_touch_send_frame(grab->seat);
+}
+
 static void default_touch_cancel(struct wlr_seat_touch_grab *grab) {
 	// cannot be cancelled
 }
 
+static void default_touch_wl_cancel(struct wlr_seat_touch_grab *grab,
+		struct wlr_surface *surface) {
+	wlr_seat_touch_send_cancel(grab->seat, surface);
+}
+
 const struct wlr_touch_grab_interface default_touch_grab_impl = {
 	.down = default_touch_down,
 	.up = default_touch_up,
 	.motion = default_touch_motion,
 	.enter = default_touch_enter,
+	.frame = default_touch_frame,
 	.cancel = default_touch_cancel,
+	.wl_cancel = default_touch_wl_cancel,
 };
 
 
@@ -73,7 +83,7 @@ void wlr_seat_touch_start_grab(struct wlr_seat *wlr_seat,
 	grab->seat = wlr_seat;
 	wlr_seat->touch_state.grab = grab;
 
-	wlr_signal_emit_safe(&wlr_seat->events.touch_grab_begin, grab);
+	wl_signal_emit_mutable(&wlr_seat->events.touch_grab_begin, grab);
 }
 
 void wlr_seat_touch_end_grab(struct wlr_seat *wlr_seat) {
@@ -81,7 +91,7 @@ void wlr_seat_touch_end_grab(struct wlr_seat *wlr_seat) {
 
 	if (grab != wlr_seat->touch_state.default_grab) {
 		wlr_seat->touch_state.grab = wlr_seat->touch_state.default_grab;
-		wlr_signal_emit_safe(&wlr_seat->events.touch_grab_end, grab);
+		wl_signal_emit_mutable(&wlr_seat->events.touch_grab_end, grab);
 		if (grab->interface->cancel) {
 			grab->interface->cancel(grab);
 		}
@@ -97,7 +107,7 @@ static void touch_point_clear_focus(struct wlr_touch_point *point) {
 }
 
 static void touch_point_destroy(struct wlr_touch_point *point) {
-	wlr_signal_emit_safe(&point->events.destroy, point);
+	wl_signal_emit_mutable(&point->events.destroy, point);
 
 	touch_point_clear_focus(point);
 	wl_list_remove(&point->surface_destroy.link);
@@ -226,6 +236,33 @@ void wlr_seat_touch_notify_motion(struct wlr_seat *seat, uint32_t time,
 	grab->interface->motion(grab, time, point);
 }
 
+void wlr_seat_touch_notify_frame(struct wlr_seat *seat) {
+	struct wlr_seat_touch_grab *grab = seat->touch_state.grab;
+	if (grab->interface->frame) {
+		grab->interface->frame(grab);
+	}
+}
+
+void wlr_seat_touch_notify_cancel(struct wlr_seat *seat,
+		struct wlr_surface *surface) {
+	struct wlr_seat_touch_grab *grab = seat->touch_state.grab;
+	if (grab->interface->wl_cancel) {
+		grab->interface->wl_cancel(grab, surface);
+	}
+
+	struct wl_client *client = wl_resource_get_client(surface->resource);
+	struct wlr_seat_client *seat_client = wlr_seat_client_for_wl_client(seat, client);
+	if (seat_client == NULL) {
+		return;
+	}
+	struct wlr_touch_point *point, *tmp;
+	wl_list_for_each_safe(point, tmp, &seat->touch_state.touch_points, link) {
+		if (point->client == seat_client) {
+			touch_point_destroy(point);
+		}
+	}
+}
+
 static void handle_point_focus_destroy(struct wl_listener *listener,
 		void *data) {
 	struct wlr_touch_point *point =
@@ -303,9 +340,10 @@ uint32_t wlr_seat_touch_send_down(struct wlr_seat *seat,
 		}
 		wl_touch_send_down(resource, serial, time, surface->resource,
 			touch_id, wl_fixed_from_double(sx), wl_fixed_from_double(sy));
-		wl_touch_send_frame(resource);
 	}
 
+	point->client->needs_touch_frame = true;
+
 	return serial;
 }
 
@@ -323,8 +361,9 @@ void wlr_seat_touch_send_up(struct wlr_seat *seat, uint32_t time, int32_t touch_
 			continue;
 		}
 		wl_touch_send_up(resource, serial, time, touch_id);
-		wl_touch_send_frame(resource);
 	}
+
+	point->client->needs_touch_frame = true;
 }
 
 void wlr_seat_touch_send_motion(struct wlr_seat *seat, uint32_t time, int32_t touch_id,
@@ -342,7 +381,39 @@ void wlr_seat_touch_send_motion(struct wlr_seat *seat, uint32_t time, int32_t to
 		}
 		wl_touch_send_motion(resource, time, touch_id, wl_fixed_from_double(sx),
 			wl_fixed_from_double(sy));
-		wl_touch_send_frame(resource);
+	}
+
+	point->client->needs_touch_frame = true;
+}
+
+void wlr_seat_touch_send_frame(struct wlr_seat *seat) {
+	struct wlr_seat_client *seat_client;
+	wl_list_for_each(seat_client, &seat->clients, link) {
+		if (!seat_client->needs_touch_frame) {
+			continue;
+		}
+
+		struct wl_resource *resource;
+		wl_resource_for_each(resource, &seat_client->touches) {
+			wl_touch_send_frame(resource);
+		}
+		seat_client->needs_touch_frame = false;
+	}
+}
+
+void wlr_seat_touch_send_cancel(struct wlr_seat *seat, struct wlr_surface *surface) {
+	struct wl_client *client = wl_resource_get_client(surface->resource);
+	struct wlr_seat_client *seat_client = wlr_seat_client_for_wl_client(seat, client);
+	if (seat_client == NULL) {
+		return;
+	}
+
+	struct wl_resource *resource;
+	wl_resource_for_each(resource, &seat_client->touches) {
+		if (seat_client_from_touch_resource(resource) == NULL) {
+			continue;
+		}
+		wl_touch_send_cancel(resource);
 	}
 }
 
