@@ -2,16 +2,18 @@
 #define _POSIX_C_SOURCE 200809L
 #endif
 
-#include "tablet-unstable-v2-protocol.h"
 #include <assert.h>
 #include <stdlib.h>
 #include <types/wlr_tablet_v2.h>
 #include <wayland-util.h>
+#include <wlr/types/wlr_compositor.h>
 #include <wlr/types/wlr_tablet_tool.h>
+#include <wlr/types/wlr_tablet_pad.h>
 #include <wlr/types/wlr_tablet_v2.h>
 #include <wlr/util/log.h>
+#include "tablet-unstable-v2-protocol.h"
 
-static struct wlr_tablet_pad_v2_grab_interface default_pad_grab_interface;
+static const struct wlr_tablet_pad_v2_grab_interface default_pad_grab_interface;
 
 struct tablet_pad_auxiliary_user_data {
 	struct wlr_tablet_pad_client_v2 *pad;
@@ -47,9 +49,9 @@ static void handle_tablet_pad_ring_v2_set_feedback(struct wl_client *client,
 		.serial = serial,
 		.description = description,
 		.index = aux->index
-		};
+	};
 
-	wl_signal_emit(&aux->pad->pad->events.ring_feedback, &evt);
+	wl_signal_emit_mutable(&aux->pad->pad->events.ring_feedback, &evt);
 }
 
 static void handle_tablet_pad_ring_v2_destroy(struct wl_client *client,
@@ -57,7 +59,7 @@ static void handle_tablet_pad_ring_v2_destroy(struct wl_client *client,
 	wl_resource_destroy(resource);
 }
 
-static struct zwp_tablet_pad_ring_v2_interface tablet_pad_ring_impl = {
+static const struct zwp_tablet_pad_ring_v2_interface tablet_pad_ring_impl = {
 	.set_feedback = handle_tablet_pad_ring_v2_set_feedback,
 	.destroy = handle_tablet_pad_ring_v2_destroy,
 };
@@ -85,9 +87,9 @@ static void handle_tablet_pad_strip_v2_set_feedback(struct wl_client *client,
 		.serial = serial,
 		.description = description,
 		.index = aux->index
-		};
+	};
 
-	wl_signal_emit(&aux->pad->pad->events.strip_feedback, &evt);
+	wl_signal_emit_mutable(&aux->pad->pad->events.strip_feedback, &evt);
 }
 
 static void handle_tablet_pad_strip_v2_destroy(struct wl_client *client,
@@ -95,7 +97,7 @@ static void handle_tablet_pad_strip_v2_destroy(struct wl_client *client,
 	wl_resource_destroy(resource);
 }
 
-static struct zwp_tablet_pad_strip_v2_interface tablet_pad_strip_impl = {
+static const struct zwp_tablet_pad_strip_v2_interface tablet_pad_strip_impl = {
 	.set_feedback = handle_tablet_pad_strip_v2_set_feedback,
 	.destroy = handle_tablet_pad_strip_v2_destroy,
 };
@@ -112,12 +114,12 @@ static void handle_tablet_pad_v2_set_feedback( struct wl_client *client,
 		.serial = serial,
 		.index = button,
 		.description = description,
-		};
+	};
 
-	wl_signal_emit(&pad->pad->events.button_feedback, &evt);
+	wl_signal_emit_mutable(&pad->pad->events.button_feedback, &evt);
 }
 
-static struct zwp_tablet_pad_v2_interface tablet_pad_impl = {
+static const struct zwp_tablet_pad_v2_interface tablet_pad_impl = {
 	.set_feedback = handle_tablet_pad_v2_set_feedback,
 	.destroy = handle_tablet_pad_v2_destroy,
 };
@@ -183,7 +185,7 @@ static void handle_tablet_pad_group_v2_destroy(struct wl_client *client,
 	wl_resource_destroy(resource);
 }
 
-static struct zwp_tablet_pad_group_v2_interface tablet_pad_group_impl = {
+static const struct zwp_tablet_pad_group_v2_interface tablet_pad_group_impl = {
 	.destroy = handle_tablet_pad_group_v2_destroy,
 };
 
@@ -305,8 +307,9 @@ void add_tablet_pad_client(struct wlr_tablet_seat_client_v2 *seat,
 		return;
 	}
 
-	client->resource =
-		wl_resource_create(seat->wl_client, &zwp_tablet_pad_v2_interface, 1, 0);
+	uint32_t version = wl_resource_get_version(seat->resource);
+	client->resource = wl_resource_create(seat->wl_client,
+		&zwp_tablet_pad_v2_interface, version, 0);
 	if (!client->resource) {
 		wl_client_post_no_memory(seat->wl_client);
 		free(client->groups);
@@ -324,10 +327,12 @@ void add_tablet_pad_client(struct wlr_tablet_seat_client_v2 *seat,
 	if (pad->wlr_pad->button_count) {
 		zwp_tablet_pad_v2_send_buttons(client->resource, pad->wlr_pad->button_count);
 	}
-	for (size_t i = 0; i < pad->wlr_pad->paths.length; ++i) {
-		zwp_tablet_pad_v2_send_path(client->resource,
-			pad->wlr_pad->paths.items[i]);
+
+	const char **path_ptr;
+	wl_array_for_each(path_ptr, &pad->wlr_pad->paths) {
+		zwp_tablet_pad_v2_send_path(client->resource, *path_ptr);
 	}
+
 	size_t i = 0;
 	struct wlr_tablet_pad_group *group;
 	client->group_count = pad->group_count;
@@ -370,7 +375,7 @@ struct wlr_tablet_v2_tablet_pad *wlr_tablet_pad_create(
 	if (!seat) {
 		return NULL;
 	}
-	struct wlr_tablet_pad *wlr_pad = wlr_device->tablet_pad;
+	struct wlr_tablet_pad *wlr_pad = wlr_tablet_pad_from_input_device(wlr_device);
 	struct wlr_tablet_v2_tablet_pad *pad = calloc(1, sizeof(struct wlr_tablet_v2_tablet_pad));
 	if (!pad) {
 		return NULL;
@@ -521,7 +526,7 @@ void wlr_send_tablet_v2_tablet_pad_ring(struct wlr_tablet_v2_tablet_pad *pad,
 	if (position < 0) {
 		zwp_tablet_pad_ring_v2_send_stop(resource);
 	} else {
-		zwp_tablet_pad_ring_v2_send_angle(resource, position);
+		zwp_tablet_pad_ring_v2_send_angle(resource, wl_fixed_from_double(position));
 	}
 	zwp_tablet_pad_ring_v2_send_frame(resource, time);
 }
@@ -695,7 +700,7 @@ static void default_pad_cancel(struct wlr_tablet_pad_v2_grab *grab) {
 	// Do nothing, the default cancel can be ignored.
 }
 
-static struct wlr_tablet_pad_v2_grab_interface default_pad_grab_interface  = {
+static const struct wlr_tablet_pad_v2_grab_interface default_pad_grab_interface  = {
 	.enter = default_pad_enter,
 	.button = default_pad_button,
 	.strip = default_pad_strip,
