@@ -7,15 +7,18 @@
 #include <wlr/types/wlr_input_device.h>
 #include <wlr/types/wlr_output_layout.h>
 #include <wlr/types/wlr_output.h>
+#include <wlr/types/wlr_pointer.h>
+#include <wlr/types/wlr_tablet_tool.h>
+#include <wlr/types/wlr_touch.h>
+#include <wlr/util/box.h>
 #include <wlr/util/log.h>
-#include "util/signal.h"
 
 struct wlr_cursor_device {
 	struct wlr_cursor *cursor;
 	struct wlr_input_device *device;
 	struct wl_list link;
 	struct wlr_output *mapped_output;
-	struct wlr_box *mapped_box;
+	struct wlr_box mapped_box; // empty if unset
 
 	struct wl_listener motion;
 	struct wl_listener motion_absolute;
@@ -28,11 +31,14 @@ struct wlr_cursor_device {
 	struct wl_listener pinch_begin;
 	struct wl_listener pinch_update;
 	struct wl_listener pinch_end;
+	struct wl_listener hold_begin;
+	struct wl_listener hold_end;
 
 	struct wl_listener touch_down;
 	struct wl_listener touch_up;
 	struct wl_listener touch_motion;
 	struct wl_listener touch_cancel;
+	struct wl_listener touch_frame;
 
 	struct wl_listener tablet_tool_axis;
 	struct wl_listener tablet_tool_proximity;
@@ -56,7 +62,7 @@ struct wlr_cursor_state {
 	struct wl_list output_cursors; // wlr_cursor_output_cursor::link
 	struct wlr_output_layout *layout;
 	struct wlr_output *mapped_output;
-	struct wlr_box *mapped_box;
+	struct wlr_box mapped_box; // empty if unset
 
 	struct wl_listener layout_add;
 	struct wl_listener layout_change;
@@ -95,12 +101,15 @@ struct wlr_cursor *wlr_cursor_create(void) {
 	wl_signal_init(&cur->events.pinch_begin);
 	wl_signal_init(&cur->events.pinch_update);
 	wl_signal_init(&cur->events.pinch_end);
+	wl_signal_init(&cur->events.hold_begin);
+	wl_signal_init(&cur->events.hold_end);
 
 	// touch signals
 	wl_signal_init(&cur->events.touch_up);
 	wl_signal_init(&cur->events.touch_down);
 	wl_signal_init(&cur->events.touch_motion);
 	wl_signal_init(&cur->events.touch_cancel);
+	wl_signal_init(&cur->events.touch_frame);
 
 	// tablet tool signals
 	wl_signal_init(&cur->events.tablet_tool_tip);
@@ -154,11 +163,14 @@ static void cursor_device_destroy(struct wlr_cursor_device *c_device) {
 		wl_list_remove(&c_device->pinch_begin.link);
 		wl_list_remove(&c_device->pinch_update.link);
 		wl_list_remove(&c_device->pinch_end.link);
+		wl_list_remove(&c_device->hold_begin.link);
+		wl_list_remove(&c_device->hold_end.link);
 	} else if (dev->type == WLR_INPUT_DEVICE_TOUCH) {
 		wl_list_remove(&c_device->touch_down.link);
 		wl_list_remove(&c_device->touch_up.link);
 		wl_list_remove(&c_device->touch_motion.link);
 		wl_list_remove(&c_device->touch_cancel.link);
+		wl_list_remove(&c_device->touch_frame.link);
 	} else if (dev->type == WLR_INPUT_DEVICE_TABLET_TOOL) {
 		wl_list_remove(&c_device->tablet_tool_axis.link);
 		wl_list_remove(&c_device->tablet_tool_proximity.link);
@@ -199,6 +211,10 @@ static struct wlr_cursor_device *get_cursor_device(struct wlr_cursor *cur,
 static void cursor_warp_unchecked(struct wlr_cursor *cur,
 		double lx, double ly) {
 	assert(cur->state->layout);
+	if (!isfinite(lx) || !isfinite(ly)) {
+		assert(false);
+		return;
+	}
 
 	struct wlr_cursor_output_cursor *output_cursor;
 	wl_list_for_each(output_cursor, &cur->state->output_cursors, link) {
@@ -224,33 +240,37 @@ static void cursor_warp_unchecked(struct wlr_cursor *cur,
  * Absolute movement for touch and pen devices will be relative to this box and
  * pointer movement will be constrained to this box.
  *
- * If none of these are set, returns NULL and absolute movement should be
+ * If none of these are set, the box is empty and absolute movement should be
  * relative to the extents of the layout.
  */
-static struct wlr_box *get_mapping(struct wlr_cursor *cur,
-		struct wlr_input_device *dev) {
+static void get_mapping(struct wlr_cursor *cur,
+		struct wlr_input_device *dev, struct wlr_box *box) {
 	assert(cur->state->layout);
-	struct wlr_cursor_device *c_device = get_cursor_device(cur, dev);
 
+	memset(box, 0, sizeof(*box));
+
+	struct wlr_cursor_device *c_device = get_cursor_device(cur, dev);
 	if (c_device) {
-		if (c_device->mapped_box) {
-			return c_device->mapped_box;
+		if (!wlr_box_empty(&c_device->mapped_box)) {
+			*box = c_device->mapped_box;
+			return;
 		}
 		if (c_device->mapped_output) {
-			return wlr_output_layout_get_box(cur->state->layout,
-				c_device->mapped_output);
+			wlr_output_layout_get_box(cur->state->layout,
+				c_device->mapped_output, box);
+			return;
 		}
 	}
 
-	if (cur->state->mapped_box) {
-		return cur->state->mapped_box;
+	if (!wlr_box_empty(&cur->state->mapped_box)) {
+		*box = cur->state->mapped_box;
+		return;
 	}
 	if (cur->state->mapped_output) {
-		return wlr_output_layout_get_box(cur->state->layout,
-			cur->state->mapped_output);
+		wlr_output_layout_get_box(cur->state->layout,
+			cur->state->mapped_output, box);
+		return;
 	}
-
-	return NULL;
 }
 
 bool wlr_cursor_warp(struct wlr_cursor *cur, struct wlr_input_device *dev,
@@ -258,9 +278,10 @@ bool wlr_cursor_warp(struct wlr_cursor *cur, struct wlr_input_device *dev,
 	assert(cur->state->layout);
 
 	bool result = false;
-	struct wlr_box *mapping = get_mapping(cur, dev);
-	if (mapping) {
-		result = wlr_box_contains_point(mapping, lx, ly);
+	struct wlr_box mapping;
+	get_mapping(cur, dev, &mapping);
+	if (!wlr_box_empty(&mapping)) {
+		result = wlr_box_contains_point(&mapping, lx, ly);
 	} else {
 		result = wlr_output_layout_contains_point(cur->state->layout, NULL,
 			lx, ly);
@@ -275,16 +296,23 @@ bool wlr_cursor_warp(struct wlr_cursor *cur, struct wlr_input_device *dev,
 
 void wlr_cursor_warp_closest(struct wlr_cursor *cur,
 		struct wlr_input_device *dev, double lx, double ly) {
-	struct wlr_box *mapping = get_mapping(cur, dev);
-	if (mapping) {
-		wlr_box_closest_point(mapping, lx, ly, &lx, &ly);
-		if (isnan(lx) || isnan(ly)) {
-			lx = 0;
-			ly = 0;
-		}
-	} else {
+	struct wlr_box mapping;
+	get_mapping(cur, dev, &mapping);
+	if (!wlr_box_empty(&mapping)) {
+		wlr_box_closest_point(&mapping, lx, ly, &lx, &ly);
+	} else if (!wl_list_empty(&cur->state->layout->outputs)) {
 		wlr_output_layout_closest_point(cur->state->layout, NULL, lx, ly,
 			&lx, &ly);
+	} else {
+		/*
+		 * There is no mapping box for the input device and the
+		 * output layout is empty.  This can happen for example
+		 * when external monitors are turned off/disconnected.
+		 * In this case, all (x,y) points are equally invalid,
+		 * so leave the cursor in its current location (better
+		 * from a user standpoint than warping it to (0,0)).
+		 */
+		return;
 	}
 
 	cursor_warp_unchecked(cur, lx, ly);
@@ -295,13 +323,14 @@ void wlr_cursor_absolute_to_layout_coords(struct wlr_cursor *cur,
 		double *lx, double *ly) {
 	assert(cur->state->layout);
 
-	struct wlr_box *mapping = get_mapping(cur, dev);
-	if (!mapping) {
-		mapping = wlr_output_layout_get_box(cur->state->layout, NULL);
+	struct wlr_box mapping;
+	get_mapping(cur, dev, &mapping);
+	if (wlr_box_empty(&mapping)) {
+		wlr_output_layout_get_box(cur->state->layout, NULL, &mapping);
 	}
 
-	*lx = !isnan(x) ? mapping->width * x + mapping->x : cur->x;
-	*ly = !isnan(y) ? mapping->height * y + mapping->y : cur->y;
+	*lx = !isnan(x) ? mapping.width * x + mapping.x : cur->x;
+	*ly = !isnan(y) ? mapping.height * y + mapping.y : cur->y;
 }
 
 void wlr_cursor_warp_absolute(struct wlr_cursor *cur,
@@ -349,10 +378,10 @@ void wlr_cursor_set_surface(struct wlr_cursor *cur, struct wlr_surface *surface,
 }
 
 static void handle_pointer_motion(struct wl_listener *listener, void *data) {
-	struct wlr_event_pointer_motion *event = data;
+	struct wlr_pointer_motion_event *event = data;
 	struct wlr_cursor_device *device =
 		wl_container_of(listener, device, motion);
-	wlr_signal_emit_safe(&device->cursor->events.motion, event);
+	wl_signal_emit_mutable(&device->cursor->events.motion, event);
 }
 
 static void apply_output_transform(double *x, double *y,
@@ -415,7 +444,7 @@ static struct wlr_output *get_mapped_output(struct wlr_cursor_device *cursor_dev
 
 static void handle_pointer_motion_absolute(struct wl_listener *listener,
 		void *data) {
-	struct wlr_event_pointer_motion_absolute *event = data;
+	struct wlr_pointer_motion_absolute_event *event = data;
 	struct wlr_cursor_device *device =
 		wl_container_of(listener, device, motion_absolute);
 
@@ -424,72 +453,84 @@ static void handle_pointer_motion_absolute(struct wl_listener *listener,
 	if (output) {
 		apply_output_transform(&event->x, &event->y, output->transform);
 	}
-	wlr_signal_emit_safe(&device->cursor->events.motion_absolute, event);
+	wl_signal_emit_mutable(&device->cursor->events.motion_absolute, event);
 }
 
 static void handle_pointer_button(struct wl_listener *listener, void *data) {
-	struct wlr_event_pointer_button *event = data;
+	struct wlr_pointer_button_event *event = data;
 	struct wlr_cursor_device *device =
 		wl_container_of(listener, device, button);
-	wlr_signal_emit_safe(&device->cursor->events.button, event);
+	wl_signal_emit_mutable(&device->cursor->events.button, event);
 }
 
 static void handle_pointer_axis(struct wl_listener *listener, void *data) {
-	struct wlr_event_pointer_axis *event = data;
+	struct wlr_pointer_axis_event *event = data;
 	struct wlr_cursor_device *device = wl_container_of(listener, device, axis);
-	wlr_signal_emit_safe(&device->cursor->events.axis, event);
+	wl_signal_emit_mutable(&device->cursor->events.axis, event);
 }
 
 static void handle_pointer_frame(struct wl_listener *listener, void *data) {
 	struct wlr_cursor_device *device = wl_container_of(listener, device, frame);
-	wlr_signal_emit_safe(&device->cursor->events.frame, device->cursor);
+	wl_signal_emit_mutable(&device->cursor->events.frame, device->cursor);
 }
 
 static void handle_pointer_swipe_begin(struct wl_listener *listener, void *data) {
-	struct wlr_event_pointer_swipe_begin *event = data;
+	struct wlr_pointer_swipe_begin_event *event = data;
 	struct wlr_cursor_device *device = wl_container_of(listener, device, swipe_begin);
-	wlr_signal_emit_safe(&device->cursor->events.swipe_begin, event);
+	wl_signal_emit_mutable(&device->cursor->events.swipe_begin, event);
 }
 
 static void handle_pointer_swipe_update(struct wl_listener *listener, void *data) {
-	struct wlr_event_pointer_swipe_update *event = data;
+	struct wlr_pointer_swipe_update_event *event = data;
 	struct wlr_cursor_device *device = wl_container_of(listener, device, swipe_update);
-	wlr_signal_emit_safe(&device->cursor->events.swipe_update, event);
+	wl_signal_emit_mutable(&device->cursor->events.swipe_update, event);
 }
 
 static void handle_pointer_swipe_end(struct wl_listener *listener, void *data) {
-	struct wlr_event_pointer_swipe_end *event = data;
+	struct wlr_pointer_swipe_end_event *event = data;
 	struct wlr_cursor_device *device = wl_container_of(listener, device, swipe_end);
-	wlr_signal_emit_safe(&device->cursor->events.swipe_end, event);
+	wl_signal_emit_mutable(&device->cursor->events.swipe_end, event);
 }
 
 static void handle_pointer_pinch_begin(struct wl_listener *listener, void *data) {
-	struct wlr_event_pointer_pinch_begin *event = data;
+	struct wlr_pointer_pinch_begin_event *event = data;
 	struct wlr_cursor_device *device = wl_container_of(listener, device, pinch_begin);
-	wlr_signal_emit_safe(&device->cursor->events.pinch_begin, event);
+	wl_signal_emit_mutable(&device->cursor->events.pinch_begin, event);
 }
 
 static void handle_pointer_pinch_update(struct wl_listener *listener, void *data) {
-	struct wlr_event_pointer_pinch_update *event = data;
+	struct wlr_pointer_pinch_update_event *event = data;
 	struct wlr_cursor_device *device = wl_container_of(listener, device, pinch_update);
-	wlr_signal_emit_safe(&device->cursor->events.pinch_update, event);
+	wl_signal_emit_mutable(&device->cursor->events.pinch_update, event);
 }
 
 static void handle_pointer_pinch_end(struct wl_listener *listener, void *data) {
-	struct wlr_event_pointer_pinch_end *event = data;
+	struct wlr_pointer_pinch_end_event *event = data;
 	struct wlr_cursor_device *device = wl_container_of(listener, device, pinch_end);
-	wlr_signal_emit_safe(&device->cursor->events.pinch_end, event);
+	wl_signal_emit_mutable(&device->cursor->events.pinch_end, event);
+}
+
+static void handle_pointer_hold_begin(struct wl_listener *listener, void *data) {
+	struct wlr_pointer_hold_begin_event *event = data;
+	struct wlr_cursor_device *device = wl_container_of(listener, device, hold_begin);
+	wl_signal_emit_mutable(&device->cursor->events.hold_begin, event);
+}
+
+static void handle_pointer_hold_end(struct wl_listener *listener, void *data) {
+	struct wlr_pointer_hold_end_event *event = data;
+	struct wlr_cursor_device *device = wl_container_of(listener, device, hold_end);
+	wl_signal_emit_mutable(&device->cursor->events.hold_end, event);
 }
 
 static void handle_touch_up(struct wl_listener *listener, void *data) {
-	struct wlr_event_touch_up *event = data;
+	struct wlr_touch_up_event *event = data;
 	struct wlr_cursor_device *device;
 	device = wl_container_of(listener, device, touch_up);
-	wlr_signal_emit_safe(&device->cursor->events.touch_up, event);
+	wl_signal_emit_mutable(&device->cursor->events.touch_up, event);
 }
 
 static void handle_touch_down(struct wl_listener *listener, void *data) {
-	struct wlr_event_touch_down *event = data;
+	struct wlr_touch_down_event *event = data;
 	struct wlr_cursor_device *device;
 	device = wl_container_of(listener, device, touch_down);
 
@@ -498,11 +539,11 @@ static void handle_touch_down(struct wl_listener *listener, void *data) {
 	if (output) {
 		apply_output_transform(&event->x, &event->y, output->transform);
 	}
-	wlr_signal_emit_safe(&device->cursor->events.touch_down, event);
+	wl_signal_emit_mutable(&device->cursor->events.touch_down, event);
 }
 
 static void handle_touch_motion(struct wl_listener *listener, void *data) {
-	struct wlr_event_touch_motion *event = data;
+	struct wlr_touch_motion_event *event = data;
 	struct wlr_cursor_device *device;
 	device = wl_container_of(listener, device, touch_motion);
 
@@ -511,18 +552,24 @@ static void handle_touch_motion(struct wl_listener *listener, void *data) {
 	if (output) {
 		apply_output_transform(&event->x, &event->y, output->transform);
 	}
-	wlr_signal_emit_safe(&device->cursor->events.touch_motion, event);
+	wl_signal_emit_mutable(&device->cursor->events.touch_motion, event);
 }
 
 static void handle_touch_cancel(struct wl_listener *listener, void *data) {
-	struct wlr_event_touch_cancel *event = data;
+	struct wlr_touch_cancel_event *event = data;
 	struct wlr_cursor_device *device;
 	device = wl_container_of(listener, device, touch_cancel);
-	wlr_signal_emit_safe(&device->cursor->events.touch_cancel, event);
+	wl_signal_emit_mutable(&device->cursor->events.touch_cancel, event);
+}
+
+static void handle_touch_frame(struct wl_listener *listener, void *data) {
+	struct wlr_cursor_device *device =
+		wl_container_of(listener, device, touch_frame);
+	wl_signal_emit_mutable(&device->cursor->events.touch_frame, NULL);
 }
 
 static void handle_tablet_tool_tip(struct wl_listener *listener, void *data) {
-	struct wlr_event_tablet_tool_tip *event = data;
+	struct wlr_tablet_tool_tip_event *event = data;
 	struct wlr_cursor_device *device;
 	device = wl_container_of(listener, device, tablet_tool_tip);
 
@@ -531,11 +578,11 @@ static void handle_tablet_tool_tip(struct wl_listener *listener, void *data) {
 	if (output) {
 		apply_output_transform(&event->x, &event->y, output->transform);
 	}
-	wlr_signal_emit_safe(&device->cursor->events.tablet_tool_tip, event);
+	wl_signal_emit_mutable(&device->cursor->events.tablet_tool_tip, event);
 }
 
 static void handle_tablet_tool_axis(struct wl_listener *listener, void *data) {
-	struct wlr_event_tablet_tool_axis *event = data;
+	struct wlr_tablet_tool_axis_event *event = data;
 	struct wlr_cursor_device *device;
 	device = wl_container_of(listener, device, tablet_tool_axis);
 
@@ -566,20 +613,20 @@ static void handle_tablet_tool_axis(struct wl_listener *listener, void *data) {
 		}
 	}
 
-	wlr_signal_emit_safe(&device->cursor->events.tablet_tool_axis, event);
+	wl_signal_emit_mutable(&device->cursor->events.tablet_tool_axis, event);
 }
 
 static void handle_tablet_tool_button(struct wl_listener *listener,
 		void *data) {
-	struct wlr_event_tablet_tool_button *event = data;
+	struct wlr_tablet_tool_button *event = data;
 	struct wlr_cursor_device *device;
 	device = wl_container_of(listener, device, tablet_tool_button);
-	wlr_signal_emit_safe(&device->cursor->events.tablet_tool_button, event);
+	wl_signal_emit_mutable(&device->cursor->events.tablet_tool_button, event);
 }
 
 static void handle_tablet_tool_proximity(struct wl_listener *listener,
 		void *data) {
-	struct wlr_event_tablet_tool_proximity *event = data;
+	struct wlr_tablet_tool_proximity_event *event = data;
 	struct wlr_cursor_device *device;
 	device = wl_container_of(listener, device, tablet_tool_proximity);
 
@@ -588,7 +635,7 @@ static void handle_tablet_tool_proximity(struct wl_listener *listener,
 	if (output) {
 		apply_output_transform(&event->x, &event->y, output->transform);
 	}
-	wlr_signal_emit_safe(&device->cursor->events.tablet_tool_proximity, event);
+	wl_signal_emit_mutable(&device->cursor->events.tablet_tool_proximity, event);
 }
 
 static void handle_device_destroy(struct wl_listener *listener, void *data) {
@@ -614,66 +661,78 @@ static struct wlr_cursor_device *cursor_device_create(
 	c_device->destroy.notify = handle_device_destroy;
 
 	if (device->type == WLR_INPUT_DEVICE_POINTER) {
-		wl_signal_add(&device->pointer->events.motion, &c_device->motion);
+		struct wlr_pointer *pointer = wlr_pointer_from_input_device(device);
+
+		wl_signal_add(&pointer->events.motion, &c_device->motion);
 		c_device->motion.notify = handle_pointer_motion;
 
-		wl_signal_add(&device->pointer->events.motion_absolute,
+		wl_signal_add(&pointer->events.motion_absolute,
 			&c_device->motion_absolute);
 		c_device->motion_absolute.notify = handle_pointer_motion_absolute;
 
-		wl_signal_add(&device->pointer->events.button, &c_device->button);
+		wl_signal_add(&pointer->events.button, &c_device->button);
 		c_device->button.notify = handle_pointer_button;
 
-		wl_signal_add(&device->pointer->events.axis, &c_device->axis);
+		wl_signal_add(&pointer->events.axis, &c_device->axis);
 		c_device->axis.notify = handle_pointer_axis;
 
-		wl_signal_add(&device->pointer->events.frame, &c_device->frame);
+		wl_signal_add(&pointer->events.frame, &c_device->frame);
 		c_device->frame.notify = handle_pointer_frame;
 
-		wl_signal_add(&device->pointer->events.swipe_begin, &c_device->swipe_begin);
+		wl_signal_add(&pointer->events.swipe_begin, &c_device->swipe_begin);
 		c_device->swipe_begin.notify = handle_pointer_swipe_begin;
 
-		wl_signal_add(&device->pointer->events.swipe_update, &c_device->swipe_update);
+		wl_signal_add(&pointer->events.swipe_update, &c_device->swipe_update);
 		c_device->swipe_update.notify = handle_pointer_swipe_update;
 
-		wl_signal_add(&device->pointer->events.swipe_end, &c_device->swipe_end);
+		wl_signal_add(&pointer->events.swipe_end, &c_device->swipe_end);
 		c_device->swipe_end.notify = handle_pointer_swipe_end;
 
-		wl_signal_add(&device->pointer->events.pinch_begin, &c_device->pinch_begin);
+		wl_signal_add(&pointer->events.pinch_begin, &c_device->pinch_begin);
 		c_device->pinch_begin.notify = handle_pointer_pinch_begin;
 
-		wl_signal_add(&device->pointer->events.pinch_update, &c_device->pinch_update);
+		wl_signal_add(&pointer->events.pinch_update, &c_device->pinch_update);
 		c_device->pinch_update.notify = handle_pointer_pinch_update;
 
-		wl_signal_add(&device->pointer->events.pinch_end, &c_device->pinch_end);
+		wl_signal_add(&pointer->events.pinch_end, &c_device->pinch_end);
 		c_device->pinch_end.notify = handle_pointer_pinch_end;
+
+		wl_signal_add(&pointer->events.hold_begin, &c_device->hold_begin);
+		c_device->hold_begin.notify = handle_pointer_hold_begin;
+
+		wl_signal_add(&pointer->events.hold_end, &c_device->hold_end);
+		c_device->hold_end.notify = handle_pointer_hold_end;
 	} else if (device->type == WLR_INPUT_DEVICE_TOUCH) {
-		wl_signal_add(&device->touch->events.motion, &c_device->touch_motion);
+		struct wlr_touch *touch = wlr_touch_from_input_device(device);
+
+		wl_signal_add(&touch->events.motion, &c_device->touch_motion);
 		c_device->touch_motion.notify = handle_touch_motion;
 
-		wl_signal_add(&device->touch->events.down, &c_device->touch_down);
+		wl_signal_add(&touch->events.down, &c_device->touch_down);
 		c_device->touch_down.notify = handle_touch_down;
 
-		wl_signal_add(&device->touch->events.up, &c_device->touch_up);
+		wl_signal_add(&touch->events.up, &c_device->touch_up);
 		c_device->touch_up.notify = handle_touch_up;
 
-		wl_signal_add(&device->touch->events.cancel, &c_device->touch_cancel);
+		wl_signal_add(&touch->events.cancel, &c_device->touch_cancel);
 		c_device->touch_cancel.notify = handle_touch_cancel;
+
+		wl_signal_add(&touch->events.frame, &c_device->touch_frame);
+		c_device->touch_frame.notify = handle_touch_frame;
 	} else if (device->type == WLR_INPUT_DEVICE_TABLET_TOOL) {
-		wl_signal_add(&device->tablet->events.tip,
-			&c_device->tablet_tool_tip);
+		struct wlr_tablet *tablet = wlr_tablet_from_input_device(device);
+
+		wl_signal_add(&tablet->events.tip, &c_device->tablet_tool_tip);
 		c_device->tablet_tool_tip.notify = handle_tablet_tool_tip;
 
-		wl_signal_add(&device->tablet->events.proximity,
+		wl_signal_add(&tablet->events.proximity,
 			&c_device->tablet_tool_proximity);
 		c_device->tablet_tool_proximity.notify = handle_tablet_tool_proximity;
 
-		wl_signal_add(&device->tablet->events.axis,
-			&c_device->tablet_tool_axis);
+		wl_signal_add(&tablet->events.axis, &c_device->tablet_tool_axis);
 		c_device->tablet_tool_axis.notify = handle_tablet_tool_axis;
 
-		wl_signal_add(&device->tablet->events.button,
-			&c_device->tablet_tool_button);
+		wl_signal_add(&tablet->events.button, &c_device->tablet_tool_button);
 		c_device->tablet_tool_button.notify = handle_tablet_tool_button;
 	}
 
@@ -770,9 +829,9 @@ static void handle_layout_change(struct wl_listener *listener, void *data) {
 	struct wlr_output_layout *layout = data;
 
 	if (!wlr_output_layout_contains_point(layout, NULL, state->cursor->x,
-			state->cursor->y)) {
+			state->cursor->y) && !wl_list_empty(&layout->outputs)) {
 		// the output we were on has gone away so go to the closest boundary
-		// point
+		// point (unless the layout is empty; compare warp_closest())
 		double x, y;
 		wlr_output_layout_closest_point(layout, NULL, state->cursor->x,
 			state->cursor->y, &x, &y);
@@ -814,7 +873,7 @@ void wlr_cursor_map_input_to_output(struct wlr_cursor *cur,
 	struct wlr_cursor_device *c_device = get_cursor_device(cur, dev);
 	if (!c_device) {
 		wlr_log(WLR_ERROR, "Cannot map device \"%s\" to output"
-			"(not found in this cursor)", dev->name);
+			" (not found in this cursor)", dev->name);
 		return;
 	}
 
@@ -822,22 +881,21 @@ void wlr_cursor_map_input_to_output(struct wlr_cursor *cur,
 }
 
 void wlr_cursor_map_to_region(struct wlr_cursor *cur,
-		struct wlr_box *box) {
-	if (box && wlr_box_empty(box)) {
-		wlr_log(WLR_ERROR, "cannot map cursor to an empty region");
-		return;
-	}
+		const struct wlr_box *box) {
+	memset(&cur->state->mapped_box, 0, sizeof(cur->state->mapped_box));
 
-	cur->state->mapped_box = box;
+	if (box) {
+		if (wlr_box_empty(box)) {
+			wlr_log(WLR_ERROR, "cannot map cursor to an empty region");
+			return;
+		}
+		cur->state->mapped_box = *box;
+	}
 }
 
 void wlr_cursor_map_input_to_region(struct wlr_cursor *cur,
-		struct wlr_input_device *dev, struct wlr_box *box) {
-	if (box && wlr_box_empty(box)) {
-		wlr_log(WLR_ERROR, "cannot map device \"%s\" input to an empty region",
-			dev->name);
-		return;
-	}
+		struct wlr_input_device *dev, const struct wlr_box *box) {
+	memset(&cur->state->mapped_box, 0, sizeof(cur->state->mapped_box));
 
 	struct wlr_cursor_device *c_device = get_cursor_device(cur, dev);
 	if (!c_device) {
@@ -846,5 +904,13 @@ void wlr_cursor_map_input_to_region(struct wlr_cursor *cur,
 		return;
 	}
 
-	c_device->mapped_box = box;
+	if (box) {
+		if (wlr_box_empty(box)) {
+			wlr_log(WLR_ERROR,
+				"cannot map device \"%s\" input to an empty region",
+				dev->name);
+			return;
+		}
+		c_device->mapped_box = *box;
+	}
 }
