@@ -6,7 +6,6 @@
 #include <wlr/types/wlr_export_dmabuf_v1.h>
 #include <wlr/types/wlr_output.h>
 #include <wlr/util/log.h>
-#include "util/signal.h"
 #include "wlr-export-dmabuf-unstable-v1-protocol.h"
 
 #define EXPORT_DMABUF_MANAGER_VERSION 1
@@ -41,8 +40,7 @@ static void frame_destroy(struct wlr_export_dmabuf_frame_v1 *frame) {
 		}
 	}
 	wl_list_remove(&frame->link);
-	wl_list_remove(&frame->output_precommit.link);
-	wlr_dmabuf_attributes_finish(&frame->attribs);
+	wl_list_remove(&frame->output_commit.link);
 	// Make the frame resource inert
 	wl_resource_set_user_data(frame->resource, NULL);
 	free(frame);
@@ -53,18 +51,39 @@ static void frame_handle_resource_destroy(struct wl_resource *resource) {
 	frame_destroy(frame);
 }
 
-static void frame_output_handle_precommit(struct wl_listener *listener,
+static void frame_output_handle_commit(struct wl_listener *listener,
 		void *data) {
 	struct wlr_export_dmabuf_frame_v1 *frame =
-		wl_container_of(listener, frame, output_precommit);
-	struct wlr_output_event_precommit *event = data;
+		wl_container_of(listener, frame, output_commit);
+	struct wlr_output_event_commit *event = data;
 
-	if (!(event->output->pending.committed & WLR_OUTPUT_STATE_BUFFER)) {
+	if (!(event->committed & WLR_OUTPUT_STATE_BUFFER)) {
 		return;
 	}
 
-	wl_list_remove(&frame->output_precommit.link);
-	wl_list_init(&frame->output_precommit.link);
+	wl_list_remove(&frame->output_commit.link);
+	wl_list_init(&frame->output_commit.link);
+
+	struct wlr_dmabuf_attributes attribs = {0};
+	if (!wlr_buffer_get_dmabuf(event->buffer, &attribs)) {
+		zwlr_export_dmabuf_frame_v1_send_cancel(frame->resource,
+			ZWLR_EXPORT_DMABUF_FRAME_V1_CANCEL_REASON_TEMPORARY);
+		frame_destroy(frame);
+		return;
+	}
+
+	uint32_t frame_flags = ZWLR_EXPORT_DMABUF_FRAME_V1_FLAGS_TRANSIENT;
+	uint32_t mod_high = attribs.modifier >> 32;
+	uint32_t mod_low = attribs.modifier & 0xFFFFFFFF;
+	zwlr_export_dmabuf_frame_v1_send_frame(frame->resource,
+		attribs.width, attribs.height, 0, 0, 0, frame_flags,
+		attribs.format, mod_high, mod_low, attribs.n_planes);
+
+	for (int i = 0; i < attribs.n_planes; ++i) {
+		off_t size = lseek(attribs.fd[i], 0, SEEK_END);
+		zwlr_export_dmabuf_frame_v1_send_object(frame->resource, i,
+			attribs.fd[i], size, attribs.offset[i], attribs.stride[i], i);
+	}
 
 	time_t tv_sec = event->when->tv_sec;
 	uint32_t tv_sec_hi = (sizeof(tv_sec) > 4) ? tv_sec >> 32 : 0;
@@ -98,7 +117,7 @@ static void manager_handle_capture_output(struct wl_client *client,
 		return;
 	}
 	frame->manager = manager;
-	wl_list_init(&frame->output_precommit.link);
+	wl_list_init(&frame->output_commit.link);
 
 	uint32_t version = wl_resource_get_version(manager_resource);
 	frame->resource = wl_resource_create(client,
@@ -113,21 +132,13 @@ static void manager_handle_capture_output(struct wl_client *client,
 
 	wl_list_insert(&manager->frames, &frame->link);
 
-	if (output == NULL || !output->enabled || !output->impl->export_dmabuf) {
+	if (output == NULL || !output->enabled) {
 		zwlr_export_dmabuf_frame_v1_send_cancel(frame->resource,
 			ZWLR_EXPORT_DMABUF_FRAME_V1_CANCEL_REASON_PERMANENT);
 		frame_destroy(frame);
 		return;
 	}
 
-	struct wlr_dmabuf_attributes *attribs = &frame->attribs;
-	if (!wlr_output_export_dmabuf(output, attribs)) {
-		zwlr_export_dmabuf_frame_v1_send_cancel(frame->resource,
-			ZWLR_EXPORT_DMABUF_FRAME_V1_CANCEL_REASON_TEMPORARY);
-		frame_destroy(frame);
-		return;
-	}
-
 	frame->output = output;
 
 	wlr_output_lock_attach_render(frame->output, true);
@@ -136,26 +147,11 @@ static void manager_handle_capture_output(struct wl_client *client,
 		frame->cursor_locked = true;
 	}
 
-	uint32_t frame_flags = ZWLR_EXPORT_DMABUF_FRAME_V1_FLAGS_TRANSIENT;
-	uint32_t mod_high = attribs->modifier >> 32;
-	uint32_t mod_low = attribs->modifier & 0xFFFFFFFF;
-
-	zwlr_export_dmabuf_frame_v1_send_frame(frame->resource,
-		output->width, output->height, 0, 0, attribs->flags, frame_flags,
-		attribs->format, mod_high, mod_low, attribs->n_planes);
-
-	for (int i = 0; i < attribs->n_planes; ++i) {
-		off_t size = lseek(attribs->fd[i], 0, SEEK_END);
-
-		zwlr_export_dmabuf_frame_v1_send_object(frame->resource, i,
-			attribs->fd[i], size, attribs->offset[i], attribs->stride[i], i);
-	}
+	wl_list_remove(&frame->output_commit.link);
+	wl_signal_add(&output->events.commit, &frame->output_commit);
+	frame->output_commit.notify = frame_output_handle_commit;
 
 	wlr_output_schedule_frame(output);
-
-	wl_list_remove(&frame->output_precommit.link);
-	wl_signal_add(&output->events.precommit, &frame->output_precommit);
-	frame->output_precommit.notify = frame_output_handle_precommit;
 }
 
 static void manager_handle_destroy(struct wl_client *client,
@@ -185,7 +181,7 @@ static void manager_bind(struct wl_client *client, void *data, uint32_t version,
 static void handle_display_destroy(struct wl_listener *listener, void *data) {
 	struct wlr_export_dmabuf_manager_v1 *manager =
 		wl_container_of(listener, manager, display_destroy);
-	wlr_signal_emit_safe(&manager->events.destroy, manager);
+	wl_signal_emit_mutable(&manager->events.destroy, manager);
 	wl_list_remove(&manager->display_destroy.link);
 	wl_global_destroy(manager->global);
 	free(manager);
