@@ -3,14 +3,17 @@
 #include <stdlib.h>
 #include <string.h>
 #include <wayland-server-core.h>
+#include <wlr/types/wlr_compositor.h>
 #include <wlr/types/wlr_layer_shell_v1.h>
 #include <wlr/types/wlr_output.h>
-#include <wlr/types/wlr_surface.h>
 #include <wlr/types/wlr_xdg_shell.h>
 #include <wlr/util/log.h>
-#include "util/signal.h"
 #include "wlr-layer-shell-unstable-v1-protocol.h"
 
+// Note: zwlr_layer_surface_v1 becomes inert on wlr_layer_surface_v1_destroy()
+
+#define LAYER_SHELL_VERSION 4
+
 static void resource_handle_destroy(struct wl_client *client,
 		struct wl_resource *resource) {
 	wl_resource_destroy(resource);
@@ -26,7 +29,7 @@ static struct wlr_layer_shell_v1 *layer_shell_from_resource(
 	return wl_resource_get_user_data(resource);
 }
 
-static struct wlr_layer_surface_v1 *layer_surface_from_resource(
+struct wlr_layer_surface_v1 *wlr_layer_surface_v1_from_resource(
 		struct wl_resource *resource) {
 	assert(wl_resource_instance_of(resource, &zwlr_layer_surface_v1_interface,
 		&layer_surface_implementation));
@@ -56,9 +59,10 @@ static void layer_surface_configure_destroy(
 
 static void layer_surface_handle_ack_configure(struct wl_client *client,
 		struct wl_resource *resource, uint32_t serial) {
-	struct wlr_layer_surface_v1 *surface = layer_surface_from_resource(resource);
+	struct wlr_layer_surface_v1 *surface =
+		wlr_layer_surface_v1_from_resource(resource);
 
-	if (!surface || surface->closed) {
+	if (!surface) {
 		return;
 	}
 
@@ -74,7 +78,7 @@ static void layer_surface_handle_ack_configure(struct wl_client *client,
 	if (!found) {
 		wl_resource_post_error(resource,
 			ZWLR_LAYER_SURFACE_V1_ERROR_INVALID_SURFACE_STATE,
-			"wrong configure serial: %u", serial);
+			"wrong configure serial: %" PRIu32, serial);
 		return;
 	}
 	// Then remove old configures from the list
@@ -85,23 +89,33 @@ static void layer_surface_handle_ack_configure(struct wl_client *client,
 		layer_surface_configure_destroy(configure);
 	}
 
-	if (surface->acked_configure) {
-		layer_surface_configure_destroy(surface->acked_configure);
-	}
-	surface->acked_configure = configure;
-	wl_list_remove(&configure->link);
-	wl_list_init(&configure->link);
+	surface->pending.configure_serial = configure->serial;
+	surface->pending.actual_width = configure->width;
+	surface->pending.actual_height = configure->height;
+
+	surface->configured = true;
+
+	layer_surface_configure_destroy(configure);
 }
 
 static void layer_surface_handle_set_size(struct wl_client *client,
 		struct wl_resource *resource, uint32_t width, uint32_t height) {
-	struct wlr_layer_surface_v1 *surface = layer_surface_from_resource(resource);
+	struct wlr_layer_surface_v1 *surface =
+		wlr_layer_surface_v1_from_resource(resource);
 
 	if (!surface) {
 		return;
 	}
-	surface->client_pending.desired_width = width;
-	surface->client_pending.desired_height = height;
+
+	if (surface->current.desired_width == width
+			&& surface->current.desired_height == height) {
+		surface->pending.committed &= ~WLR_LAYER_SURFACE_V1_STATE_DESIRED_SIZE;
+	} else {
+		surface->pending.committed |= WLR_LAYER_SURFACE_V1_STATE_DESIRED_SIZE;
+	}
+
+	surface->pending.desired_width = width;
+	surface->pending.desired_height = height;
 }
 
 static void layer_surface_handle_set_anchor(struct wl_client *client,
@@ -114,83 +128,126 @@ static void layer_surface_handle_set_anchor(struct wl_client *client,
 	if (anchor > max_anchor) {
 		wl_resource_post_error(resource,
 			ZWLR_LAYER_SURFACE_V1_ERROR_INVALID_ANCHOR,
-			"invalid anchor %d", anchor);
+			"invalid anchor %" PRIu32, anchor);
 	}
-	struct wlr_layer_surface_v1 *surface = layer_surface_from_resource(resource);
+	struct wlr_layer_surface_v1 *surface =
+		wlr_layer_surface_v1_from_resource(resource);
 
 	if (!surface) {
 		return;
 	}
-	surface->client_pending.anchor = anchor;
+
+	if (surface->current.anchor == anchor) {
+		surface->pending.committed &= ~WLR_LAYER_SURFACE_V1_STATE_ANCHOR;
+	} else {
+		surface->pending.committed |= WLR_LAYER_SURFACE_V1_STATE_ANCHOR;
+	}
+
+	surface->pending.anchor = anchor;
 }
 
 static void layer_surface_handle_set_exclusive_zone(struct wl_client *client,
 		struct wl_resource *resource, int32_t zone) {
-	struct wlr_layer_surface_v1 *surface = layer_surface_from_resource(resource);
+	struct wlr_layer_surface_v1 *surface = wlr_layer_surface_v1_from_resource(resource);
 
 	if (!surface) {
 		return;
 	}
-	surface->client_pending.exclusive_zone = zone;
+
+	if (surface->current.exclusive_zone == zone) {
+		surface->pending.committed &= ~WLR_LAYER_SURFACE_V1_STATE_EXCLUSIVE_ZONE;
+	} else {
+		surface->pending.committed |= WLR_LAYER_SURFACE_V1_STATE_EXCLUSIVE_ZONE;
+	}
+
+	surface->pending.exclusive_zone = zone;
 }
 
 static void layer_surface_handle_set_margin(
 		struct wl_client *client, struct wl_resource *resource,
 		int32_t top, int32_t right, int32_t bottom, int32_t left) {
-	struct wlr_layer_surface_v1 *surface = layer_surface_from_resource(resource);
+	struct wlr_layer_surface_v1 *surface =
+		wlr_layer_surface_v1_from_resource(resource);
 
 	if (!surface) {
 		return;
 	}
-	surface->client_pending.margin.top = top;
-	surface->client_pending.margin.right = right;
-	surface->client_pending.margin.bottom = bottom;
-	surface->client_pending.margin.left = left;
+
+	if (surface->current.margin.top == top
+			&& surface->current.margin.right == right
+			&& surface->current.margin.bottom == bottom
+			&& surface->current.margin.left == left) {
+		surface->pending.committed &= ~WLR_LAYER_SURFACE_V1_STATE_MARGIN;
+	} else {
+		surface->pending.committed |= WLR_LAYER_SURFACE_V1_STATE_MARGIN;
+	}
+
+	surface->pending.margin.top = top;
+	surface->pending.margin.right = right;
+	surface->pending.margin.bottom = bottom;
+	surface->pending.margin.left = left;
 }
 
 static void layer_surface_handle_set_keyboard_interactivity(
 		struct wl_client *client, struct wl_resource *resource,
 		uint32_t interactive) {
-	struct wlr_layer_surface_v1 *surface = layer_surface_from_resource(resource);
+	struct wlr_layer_surface_v1 *surface = wlr_layer_surface_v1_from_resource(resource);
 
 	if (!surface) {
 		return;
 	}
-	surface->client_pending.keyboard_interactive = !!interactive;
+
+	surface->pending.committed |= WLR_LAYER_SURFACE_V1_STATE_KEYBOARD_INTERACTIVITY;
+	if (wl_resource_get_version(resource) < ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_ON_DEMAND_SINCE_VERSION) {
+		surface->pending.keyboard_interactive = !!interactive;
+	} else {
+		if (interactive > ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_ON_DEMAND) {
+			wl_resource_post_error(resource,
+				ZWLR_LAYER_SURFACE_V1_ERROR_INVALID_KEYBOARD_INTERACTIVITY,
+				"wrong keyboard interactivity value: %" PRIu32, interactive);
+		} else {
+			surface->pending.keyboard_interactive = interactive;
+		}
+	}
 }
 
 static void layer_surface_handle_get_popup(struct wl_client *client,
 		struct wl_resource *layer_resource,
 		struct wl_resource *popup_resource) {
 	struct wlr_layer_surface_v1 *parent =
-		layer_surface_from_resource(layer_resource);
-	struct wlr_xdg_surface *popup_surface =
-		wlr_xdg_surface_from_popup_resource(popup_resource);
+		wlr_layer_surface_v1_from_resource(layer_resource);
+	struct wlr_xdg_popup *popup =
+		wlr_xdg_popup_from_resource(popup_resource);
 
 	if (!parent) {
 		return;
 	}
-	assert(popup_surface->role == WLR_XDG_SURFACE_ROLE_POPUP);
-	struct wlr_xdg_popup *popup = popup_surface->popup;
 	popup->parent = parent->surface;
 	wl_list_insert(&parent->popups, &popup->link);
-	wlr_signal_emit_safe(&parent->events.new_popup, popup);
+	wl_signal_emit_mutable(&parent->events.new_popup, popup);
 }
 
 static void layer_surface_set_layer(struct wl_client *client,
 		struct wl_resource *surface_resource, uint32_t layer) {
 	struct wlr_layer_surface_v1 *surface =
-			layer_surface_from_resource(surface_resource);
+			wlr_layer_surface_v1_from_resource(surface_resource);
 	if (!surface) {
 		return;
 	}
 	if (layer > ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY) {
 		wl_resource_post_error(surface->resource,
 				ZWLR_LAYER_SHELL_V1_ERROR_INVALID_LAYER,
-				"Invalid layer %d", layer);
+				"Invalid layer %" PRIu32, layer);
 		return;
 	}
-	surface->client_pending.layer = layer;
+
+	if (surface->current.layer == layer) {
+		surface->pending.committed &= ~WLR_LAYER_SURFACE_V1_STATE_LAYER;
+	} else {
+		surface->pending.committed |= WLR_LAYER_SURFACE_V1_STATE_LAYER;
+	}
+
+	surface->pending.layer = layer;
 }
 
 static const struct zwlr_layer_surface_v1_interface layer_surface_implementation = {
@@ -206,122 +263,81 @@ static const struct zwlr_layer_surface_v1_interface layer_surface_implementation
 };
 
 static void layer_surface_unmap(struct wlr_layer_surface_v1 *surface) {
+	surface->configured = surface->mapped = false;
+
 	// TODO: probably need to ungrab before this event
-	wlr_signal_emit_safe(&surface->events.unmap, surface);
+	wl_signal_emit_mutable(&surface->events.unmap, surface);
 
 	struct wlr_xdg_popup *popup, *popup_tmp;
 	wl_list_for_each_safe(popup, popup_tmp, &surface->popups, link) {
-		wlr_xdg_popup_destroy(popup->base);
+		wlr_xdg_popup_destroy(popup);
 	}
 
 	struct wlr_layer_surface_v1_configure *configure, *tmp;
 	wl_list_for_each_safe(configure, tmp, &surface->configure_list, link) {
 		layer_surface_configure_destroy(configure);
 	}
-
-	surface->configured = surface->mapped = false;
-	surface->configure_serial = 0;
-	surface->configure_next_serial = 0;
-}
-
-static void layer_surface_destroy(struct wlr_layer_surface_v1 *surface) {
-	if (surface->configured && surface->mapped) {
-		layer_surface_unmap(surface);
-	}
-	wlr_signal_emit_safe(&surface->events.destroy, surface);
-	wl_resource_set_user_data(surface->resource, NULL);
-	surface->surface->role_data = NULL;
-	wl_list_remove(&surface->surface_destroy.link);
-	free(surface->namespace);
-	free(surface);
 }
 
 static void layer_surface_resource_destroy(struct wl_resource *resource) {
 	struct wlr_layer_surface_v1 *surface =
-		layer_surface_from_resource(resource);
+		wlr_layer_surface_v1_from_resource(resource);
 	if (surface != NULL) {
-		layer_surface_destroy(surface);
+		wlr_surface_destroy_role_object(surface->surface);
 	}
 }
 
-static bool layer_surface_state_changed(struct wlr_layer_surface_v1 *surface) {
-	struct wlr_layer_surface_v1_state *state;
-	if (wl_list_empty(&surface->configure_list)) {
-		if (surface->acked_configure) {
-			state = &surface->acked_configure->state;
-		} else if (!surface->configured) {
-			return true;
-		} else {
-			state = &surface->current;
-		}
-	} else {
-		struct wlr_layer_surface_v1_configure *configure =
-			wl_container_of(surface->configure_list.prev, configure, link);
-		state = &configure->state;
-	}
-
-	bool changed = state->actual_width != surface->server_pending.actual_width
-		|| state->actual_height != surface->server_pending.actual_height;
-	return changed;
-}
-
-void wlr_layer_surface_v1_configure(struct wlr_layer_surface_v1 *surface,
+uint32_t wlr_layer_surface_v1_configure(struct wlr_layer_surface_v1 *surface,
 		uint32_t width, uint32_t height) {
-	surface->server_pending.actual_width = width;
-	surface->server_pending.actual_height = height;
-	if (layer_surface_state_changed(surface)) {
-		struct wl_display *display =
-			wl_client_get_display(wl_resource_get_client(surface->resource));
-		struct wlr_layer_surface_v1_configure *configure =
-			calloc(1, sizeof(struct wlr_layer_surface_v1_configure));
-		if (configure == NULL) {
-			wl_client_post_no_memory(wl_resource_get_client(surface->resource));
-			return;
-		}
-		surface->configure_next_serial = wl_display_next_serial(display);
-		wl_list_insert(surface->configure_list.prev, &configure->link);
-		configure->state.actual_width = width;
-		configure->state.actual_height = height;
-		configure->serial = surface->configure_next_serial;
-		zwlr_layer_surface_v1_send_configure(surface->resource,
-				configure->serial, configure->state.actual_width,
-				configure->state.actual_height);
+	struct wl_display *display =
+		wl_client_get_display(wl_resource_get_client(surface->resource));
+	struct wlr_layer_surface_v1_configure *configure =
+		calloc(1, sizeof(struct wlr_layer_surface_v1_configure));
+	if (configure == NULL) {
+		wl_client_post_no_memory(wl_resource_get_client(surface->resource));
+		return surface->pending.configure_serial;
 	}
+	wl_list_insert(surface->configure_list.prev, &configure->link);
+	configure->width = width;
+	configure->height = height;
+	configure->serial = wl_display_next_serial(display);
+	zwlr_layer_surface_v1_send_configure(surface->resource,
+			configure->serial, configure->width,
+			configure->height);
+	return configure->serial;
 }
 
-void wlr_layer_surface_v1_close(struct wlr_layer_surface_v1 *surface) {
-	if (surface->closed) {
-		return;
-	}
-	surface->closed = true;
-	if (surface->mapped) {
-		layer_surface_unmap(surface);
-	}
+void wlr_layer_surface_v1_destroy(struct wlr_layer_surface_v1 *surface) {
 	zwlr_layer_surface_v1_send_closed(surface->resource);
+	wlr_surface_destroy_role_object(surface->surface);
 }
 
 static void layer_surface_role_commit(struct wlr_surface *wlr_surface) {
 	struct wlr_layer_surface_v1 *surface =
 		wlr_layer_surface_v1_from_wlr_surface(wlr_surface);
-	if (surface == NULL) {
+
+	const uint32_t horiz = ZWLR_LAYER_SURFACE_V1_ANCHOR_LEFT |
+		ZWLR_LAYER_SURFACE_V1_ANCHOR_RIGHT;
+	if (surface->pending.desired_width == 0 &&
+		(surface->pending.anchor & horiz) != horiz) {
+		wl_resource_post_error(surface->resource,
+			ZWLR_LAYER_SURFACE_V1_ERROR_INVALID_SIZE,
+			"width 0 requested without setting left and right anchors");
 		return;
 	}
 
-	if (surface->closed) {
-		// Ignore commits after the compositor has closed it
+	const uint32_t vert = ZWLR_LAYER_SURFACE_V1_ANCHOR_TOP |
+		ZWLR_LAYER_SURFACE_V1_ANCHOR_BOTTOM;
+	if (surface->pending.desired_height == 0 &&
+		(surface->pending.anchor & vert) != vert) {
+		wl_resource_post_error(surface->resource,
+			ZWLR_LAYER_SURFACE_V1_ERROR_INVALID_SIZE,
+			"height 0 requested without setting top and bottom anchors");
 		return;
 	}
 
-	if (surface->acked_configure) {
-		struct wlr_layer_surface_v1_configure *configure =
-			surface->acked_configure;
-		surface->configured = true;
-		surface->configure_serial = configure->serial;
-		surface->current.actual_width = configure->state.actual_width;
-		surface->current.actual_height = configure->state.actual_height;
-		layer_surface_configure_destroy(configure);
-		surface->acked_configure = NULL;
-	}
+	surface->current = surface->pending;
+	surface->pending.committed = 0;
 
 	if (wlr_surface_has_buffer(surface->surface) && !surface->configured) {
 		wl_resource_post_error(surface->resource,
@@ -330,46 +346,56 @@ static void layer_surface_role_commit(struct wlr_surface *wlr_surface) {
 		return;
 	}
 
-	surface->current.anchor = surface->client_pending.anchor;
-	surface->current.exclusive_zone = surface->client_pending.exclusive_zone;
-	surface->current.margin = surface->client_pending.margin;
-	surface->current.keyboard_interactive =
-		surface->client_pending.keyboard_interactive;
-	surface->current.desired_width = surface->client_pending.desired_width;
-	surface->current.desired_height = surface->client_pending.desired_height;
-	surface->current.layer = surface->client_pending.layer;
-
 	if (!surface->added) {
 		surface->added = true;
-		wlr_signal_emit_safe(&surface->shell->events.new_surface,
-				surface);
-		// either the compositor found a suitable output or it must
-		// have closed the surface
-		assert(surface->output || surface->closed);
+		assert(!surface->configured);
+		assert(!surface->mapped);
+		wl_signal_emit_mutable(&surface->shell->events.new_surface, surface);
+		// Return early here as the compositor may have closed this layer surface
+		// in response to the new_surface event.
+		return;
 	}
+
 	if (surface->configured && wlr_surface_has_buffer(surface->surface) &&
 			!surface->mapped) {
 		surface->mapped = true;
-		wlr_signal_emit_safe(&surface->events.map, surface);
+		wl_signal_emit_mutable(&surface->events.map, surface);
 	}
-	if (surface->configured && !wlr_surface_has_buffer(surface->surface) &&
-			surface->mapped) {
+}
+
+static void layer_surface_role_precommit(struct wlr_surface *wlr_surface,
+		const struct wlr_surface_state *state) {
+	struct wlr_layer_surface_v1 *surface =
+		wlr_layer_surface_v1_from_wlr_surface(wlr_surface);
+
+	if (state->committed & WLR_SURFACE_STATE_BUFFER && state->buffer == NULL) {
+		// This is a NULL commit
+		if (surface->configured && surface->mapped) {
+			layer_surface_unmap(surface);
+		}
+	}
+}
+
+static void layer_surface_role_destroy(struct wlr_surface *wlr_surface) {
+	struct wlr_layer_surface_v1 *surface =
+		wlr_layer_surface_v1_from_wlr_surface(wlr_surface);
+
+	if (surface->configured && surface->mapped) {
 		layer_surface_unmap(surface);
 	}
+	wl_signal_emit_mutable(&surface->events.destroy, surface);
+	wl_resource_set_user_data(surface->resource, NULL);
+	free(surface->namespace);
+	free(surface);
 }
 
 static const struct wlr_surface_role layer_surface_role = {
 	.name = "zwlr_layer_surface_v1",
 	.commit = layer_surface_role_commit,
+	.precommit = layer_surface_role_precommit,
+	.destroy = layer_surface_role_destroy,
 };
 
-static void handle_surface_destroyed(struct wl_listener *listener,
-		void *data) {
-	struct wlr_layer_surface_v1 *layer_surface =
-		wl_container_of(listener, layer_surface, surface_destroy);
-	layer_surface_destroy(layer_surface);
-}
-
 static void layer_shell_handle_get_layer_surface(struct wl_client *wl_client,
 		struct wl_resource *client_resource, uint32_t id,
 		struct wl_resource *surface_resource,
@@ -398,12 +424,12 @@ static void layer_shell_handle_get_layer_surface(struct wl_client *wl_client,
 	if (output_resource) {
 		surface->output = wlr_output_from_resource(output_resource);
 	}
-	surface->current.layer = surface->client_pending.layer = layer;
+	surface->current.layer = surface->pending.layer = layer;
 	if (layer > ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY) {
 		free(surface);
 		wl_resource_post_error(client_resource,
 				ZWLR_LAYER_SHELL_V1_ERROR_INVALID_LAYER,
-				"Invalid layer %d", layer);
+				"Invalid layer %" PRIu32, layer);
 		return;
 	}
 	surface->namespace = strdup(namespace);
@@ -431,10 +457,6 @@ static void layer_shell_handle_get_layer_surface(struct wl_client *wl_client,
 	wl_signal_init(&surface->events.unmap);
 	wl_signal_init(&surface->events.new_popup);
 
-	wl_signal_add(&surface->surface->events.destroy,
-		&surface->surface_destroy);
-	surface->surface_destroy.notify = handle_surface_destroyed;
-
 	wlr_log(WLR_DEBUG, "new layer_surface %p (res %p)",
 			surface, surface->resource);
 	wl_resource_set_implementation(surface->resource,
@@ -464,7 +486,7 @@ static void layer_shell_bind(struct wl_client *wl_client, void *data,
 static void handle_display_destroy(struct wl_listener *listener, void *data) {
 	struct wlr_layer_shell_v1 *layer_shell =
 		wl_container_of(listener, layer_shell, display_destroy);
-	wlr_signal_emit_safe(&layer_shell->events.destroy, layer_shell);
+	wl_signal_emit_mutable(&layer_shell->events.destroy, layer_shell);
 	wl_list_remove(&layer_shell->display_destroy.link);
 	wl_global_destroy(layer_shell->global);
 	free(layer_shell);
@@ -478,7 +500,8 @@ struct wlr_layer_shell_v1 *wlr_layer_shell_v1_create(struct wl_display *display)
 	}
 
 	struct wl_global *global = wl_global_create(display,
-		&zwlr_layer_shell_v1_interface, 2, layer_shell, layer_shell_bind);
+		&zwlr_layer_shell_v1_interface, LAYER_SHELL_VERSION,
+		layer_shell, layer_shell_bind);
 	if (!global) {
 		free(layer_shell);
 		return NULL;
@@ -507,82 +530,65 @@ static void layer_surface_iterator(struct wlr_surface *surface,
 		iter_data->user_data);
 }
 
-static void xdg_surface_for_each_surface(struct wlr_xdg_surface *surface,
-		int x, int y, wlr_surface_iterator_func_t iterator, void *user_data) {
-	struct layer_surface_iterator_data data = {
-		.user_iterator = iterator,
-		.user_data = user_data,
-		.x = x, .y = y,
-	};
-	wlr_surface_for_each_surface(
-			surface->surface, layer_surface_iterator, &data);
-
-	struct wlr_xdg_popup *popup_state;
-	wl_list_for_each(popup_state, &surface->popups, link) {
-		struct wlr_xdg_surface *popup = popup_state->base;
-		if (!popup->configured) {
-			continue;
-		}
-
-		double popup_sx = popup_state->geometry.x - popup_state->base->geometry.x;
-		double popup_sy = popup_state->geometry.y - popup_state->base->geometry.y;
-
-		xdg_surface_for_each_surface(popup,
-			x + popup_sx,
-			y + popup_sy,
-			iterator, user_data);
-	}
+void wlr_layer_surface_v1_for_each_surface(struct wlr_layer_surface_v1 *surface,
+		wlr_surface_iterator_func_t iterator, void *user_data) {
+	wlr_surface_for_each_surface(surface->surface, iterator, user_data);
+	wlr_layer_surface_v1_for_each_popup_surface(surface, iterator, user_data);
 }
 
-static void layer_surface_for_each_surface(struct wlr_layer_surface_v1 *surface,
-		int x, int y, wlr_surface_iterator_func_t iterator, void *user_data) {
-	struct layer_surface_iterator_data data = {
-		.user_iterator = iterator,
-		.user_data = user_data,
-		.x = x, .y = y,
-	};
-	wlr_surface_for_each_surface(surface->surface,
-			layer_surface_iterator, &data);
-
-	struct wlr_xdg_popup *popup_state;
-	wl_list_for_each(popup_state, &surface->popups, link) {
-		struct wlr_xdg_surface *popup = popup_state->base;
-		if (!popup->configured) {
+void wlr_layer_surface_v1_for_each_popup_surface(struct wlr_layer_surface_v1 *surface,
+		wlr_surface_iterator_func_t iterator, void *user_data){
+	struct wlr_xdg_popup *popup;
+	wl_list_for_each(popup, &surface->popups, link) {
+		if (!popup->base->configured || !popup->base->mapped) {
 			continue;
 		}
 
 		double popup_sx, popup_sy;
-		popup_sx = popup->popup->geometry.x - popup->geometry.x;
-		popup_sy = popup->popup->geometry.y - popup->geometry.y;
+		popup_sx = popup->current.geometry.x - popup->base->current.geometry.x;
+		popup_sy = popup->current.geometry.y - popup->base->current.geometry.y;
 
-		xdg_surface_for_each_surface(popup,
-			popup_sx, popup_sy, iterator, user_data);
+		struct layer_surface_iterator_data data = {
+			.user_iterator = iterator,
+			.user_data = user_data,
+			.x = popup_sx, .y = popup_sy,
+		};
+
+		wlr_xdg_surface_for_each_surface(popup->base,
+			layer_surface_iterator, &data);
 	}
 }
 
-void wlr_layer_surface_v1_for_each_surface(struct wlr_layer_surface_v1 *surface,
-		wlr_surface_iterator_func_t iterator, void *user_data) {
-	layer_surface_for_each_surface(surface, 0, 0, iterator, user_data);
-}
-
 struct wlr_surface *wlr_layer_surface_v1_surface_at(
 		struct wlr_layer_surface_v1 *surface, double sx, double sy,
 		double *sub_x, double *sub_y) {
-	struct wlr_xdg_popup *popup_state;
-	wl_list_for_each(popup_state, &surface->popups, link) {
-		struct wlr_xdg_surface *popup = popup_state->base;
+	struct wlr_surface *sub = wlr_layer_surface_v1_popup_surface_at(surface,
+			sx, sy, sub_x, sub_y);
+	if (sub != NULL) {
+		return sub;
+	}
+	return wlr_surface_surface_at(surface->surface, sx, sy, sub_x, sub_y);
+}
 
-		double popup_sx = popup_state->geometry.x - popup->geometry.x;
-		double popup_sy = popup_state->geometry.y - popup->geometry.y;
+struct wlr_surface *wlr_layer_surface_v1_popup_surface_at(
+		struct wlr_layer_surface_v1 *surface, double sx, double sy,
+		double *sub_x, double *sub_y) {
+	struct wlr_xdg_popup *popup;
+	wl_list_for_each(popup, &surface->popups, link) {
+		if (!popup->base->mapped) {
+			continue;
+		}
 
-		struct wlr_surface *sub = wlr_xdg_surface_surface_at(popup,
-			sx - popup_sx,
-			sy - popup_sy,
-			sub_x, sub_y);
+		double popup_sx, popup_sy;
+		popup_sx = popup->current.geometry.x - popup->base->current.geometry.x;
+		popup_sy = popup->current.geometry.y - popup->base->current.geometry.y;
+
+		struct wlr_surface *sub = wlr_xdg_surface_surface_at(
+			popup->base, sx - popup_sx, sy - popup_sy, sub_x, sub_y);
 		if (sub != NULL) {
 			return sub;
 		}
 	}
 
-	return wlr_surface_surface_at(surface->surface, sx, sy, sub_x, sub_y);
+	return NULL;
 }
