@@ -9,7 +9,6 @@
 #include <wlr/types/wlr_seat.h>
 #include <wlr/util/log.h>
 #include "primary-selection-unstable-v1-protocol.h"
-#include "util/signal.h"
 
 #define DEVICE_MANAGER_VERSION 1
 
@@ -452,7 +451,13 @@ static void primary_selection_device_manager_bind(struct wl_client *client,
 static void handle_display_destroy(struct wl_listener *listener, void *data) {
 	struct wlr_primary_selection_v1_device_manager *manager =
 		wl_container_of(listener, manager, display_destroy);
-	wlr_signal_emit_safe(&manager->events.destroy, manager);
+
+	struct wlr_primary_selection_v1_device *device, *tmp;
+	wl_list_for_each_safe(device, tmp, &manager->devices, link) {
+		device_destroy(device);
+	}
+
+	wl_signal_emit_mutable(&manager->events.destroy, manager);
 	wl_list_remove(&manager->display_destroy.link);
 	wl_global_destroy(manager->global);
 	free(manager);
