@@ -3,7 +3,6 @@
 #include <stdlib.h>
 #include <wlr/types/wlr_xdg_decoration_v1.h>
 #include <wlr/util/log.h>
-#include "util/signal.h"
 #include "xdg-decoration-unstable-v1-protocol.h"
 
 #define DECORATION_MANAGER_VERSION 1
@@ -29,9 +28,9 @@ static void toplevel_decoration_handle_set_mode(struct wl_client *client,
 	struct wlr_xdg_toplevel_decoration_v1 *decoration =
 		toplevel_decoration_from_resource(resource);
 
-	decoration->client_pending_mode =
+	decoration->requested_mode =
 		(enum wlr_xdg_toplevel_decoration_v1_mode)mode;
-	wlr_signal_emit_safe(&decoration->events.request_mode, decoration);
+	wl_signal_emit_mutable(&decoration->events.request_mode, decoration);
 }
 
 static void toplevel_decoration_handle_unset_mode(struct wl_client *client,
@@ -39,8 +38,8 @@ static void toplevel_decoration_handle_unset_mode(struct wl_client *client,
 	struct wlr_xdg_toplevel_decoration_v1 *decoration =
 		toplevel_decoration_from_resource(resource);
 
-	decoration->client_pending_mode = WLR_XDG_TOPLEVEL_DECORATION_V1_MODE_NONE;
-	wlr_signal_emit_safe(&decoration->events.request_mode, decoration);
+	decoration->requested_mode = WLR_XDG_TOPLEVEL_DECORATION_V1_MODE_NONE;
+	wl_signal_emit_mutable(&decoration->events.request_mode, decoration);
 }
 
 static const struct zxdg_toplevel_decoration_v1_interface
@@ -54,7 +53,7 @@ uint32_t wlr_xdg_toplevel_decoration_v1_set_mode(
 		struct wlr_xdg_toplevel_decoration_v1 *decoration,
 		enum wlr_xdg_toplevel_decoration_v1_mode mode) {
 	assert(mode != WLR_XDG_TOPLEVEL_DECORATION_V1_MODE_NONE);
-	decoration->server_pending_mode = mode;
+	decoration->scheduled_mode = mode;
 	return wlr_xdg_surface_schedule_configure(decoration->surface);
 }
 
@@ -62,7 +61,7 @@ static void toplevel_decoration_handle_resource_destroy(
 		struct wl_resource *resource) {
 	struct wlr_xdg_toplevel_decoration_v1 *decoration =
 		toplevel_decoration_from_resource(resource);
-	wlr_signal_emit_safe(&decoration->events.destroy, decoration);
+	wl_signal_emit_mutable(&decoration->events.destroy, decoration);
 	wl_list_remove(&decoration->surface_commit.link);
 	wl_list_remove(&decoration->surface_destroy.link);
 	wl_list_remove(&decoration->surface_configure.link);
@@ -88,7 +87,7 @@ static void toplevel_decoration_handle_surface_configure(
 		wl_container_of(listener, decoration, surface_configure);
 	struct wlr_xdg_surface_configure *surface_configure = data;
 
-	if (decoration->current_mode == decoration->server_pending_mode) {
+	if (decoration->pending.mode == decoration->scheduled_mode) {
 		return;
 	}
 
@@ -98,7 +97,7 @@ static void toplevel_decoration_handle_surface_configure(
 		return;
 	}
 	configure->surface_configure = surface_configure;
-	configure->mode = decoration->server_pending_mode;
+	configure->mode = decoration->scheduled_mode;
 	wl_list_insert(decoration->configure_list.prev, &configure->link);
 
 	zxdg_toplevel_decoration_v1_send_configure(decoration->resource,
@@ -132,7 +131,7 @@ static void toplevel_decoration_handle_surface_ack_configure(
 		free(configure);
 	}
 
-	decoration->current_mode = configure->mode;
+	decoration->pending.mode = configure->mode;
 
 	wl_list_remove(&configure->link);
 	free(configure);
@@ -144,17 +143,15 @@ static void toplevel_decoration_handle_surface_commit(
 		wl_container_of(listener, decoration, surface_commit);
 	struct wlr_xdg_decoration_manager_v1 *manager = decoration->manager;
 
-	if (decoration->surface->added) {
-		wl_list_remove(&decoration->surface_commit.link);
-		wl_list_init(&decoration->surface_commit.link);
+	decoration->current = decoration->pending;
 
+	if (decoration->surface->added && !decoration->added) {
 		decoration->added = true;
-		wlr_signal_emit_safe(&manager->events.new_toplevel_decoration,
+		wl_signal_emit_mutable(&manager->events.new_toplevel_decoration,
 			decoration);
 	}
 }
 
-
 static const struct zxdg_decoration_manager_v1_interface decoration_manager_impl;
 
 static struct wlr_xdg_decoration_manager_v1 *
@@ -175,11 +172,10 @@ static void decoration_manager_handle_get_toplevel_decoration(
 		uint32_t id, struct wl_resource *toplevel_resource) {
 	struct wlr_xdg_decoration_manager_v1 *manager =
 		decoration_manager_from_resource(manager_resource);
-	struct wlr_xdg_surface *surface =
-		wlr_xdg_surface_from_toplevel_resource(toplevel_resource);
-	assert(surface->role == WLR_XDG_SURFACE_ROLE_TOPLEVEL);
+	struct wlr_xdg_toplevel *toplevel =
+		wlr_xdg_toplevel_from_resource(toplevel_resource);
 
-	if (wlr_surface_has_buffer(surface->surface)) {
+	if (wlr_surface_has_buffer(toplevel->base->surface)) {
 		wl_resource_post_error(manager_resource,
 			ZXDG_TOPLEVEL_DECORATION_V1_ERROR_UNCONFIGURED_BUFFER,
 			"xdg_toplevel_decoration must not have a buffer at creation");
@@ -193,7 +189,7 @@ static void decoration_manager_handle_get_toplevel_decoration(
 		return;
 	}
 	decoration->manager = manager;
-	decoration->surface = surface;
+	decoration->surface = toplevel->base;
 
 	uint32_t version = wl_resource_get_version(manager_resource);
 	decoration->resource = wl_resource_create(client,
@@ -214,30 +210,29 @@ static void decoration_manager_handle_get_toplevel_decoration(
 	wl_signal_init(&decoration->events.destroy);
 	wl_signal_init(&decoration->events.request_mode);
 
-	wl_signal_add(&surface->events.destroy, &decoration->surface_destroy);
+	wl_signal_add(&toplevel->base->events.destroy,
+		&decoration->surface_destroy);
 	decoration->surface_destroy.notify =
 		toplevel_decoration_handle_surface_destroy;
-	wl_signal_add(&surface->events.configure, &decoration->surface_configure);
+	wl_signal_add(&toplevel->base->events.configure,
+		&decoration->surface_configure);
 	decoration->surface_configure.notify =
 		toplevel_decoration_handle_surface_configure;
-	wl_signal_add(&surface->events.ack_configure,
+	wl_signal_add(&toplevel->base->events.ack_configure,
 		&decoration->surface_ack_configure);
 	decoration->surface_ack_configure.notify =
 		toplevel_decoration_handle_surface_ack_configure;
-	wl_list_init(&decoration->surface_commit.link);
+	wl_signal_add(&toplevel->base->surface->events.commit,
+		&decoration->surface_commit);
+	decoration->surface_commit.notify =
+		toplevel_decoration_handle_surface_commit;
 
 	wl_list_insert(&manager->decorations, &decoration->link);
 
-	if (surface->added) {
+	if (toplevel->base->added) {
 		decoration->added = true;
-		wlr_signal_emit_safe(&manager->events.new_toplevel_decoration,
+		wl_signal_emit_mutable(&manager->events.new_toplevel_decoration,
 			decoration);
-	} else {
-		wl_list_remove(&decoration->surface_commit.link);
-		wl_signal_add(&surface->surface->events.commit,
-			&decoration->surface_commit);
-		decoration->surface_commit.notify =
-			toplevel_decoration_handle_surface_commit;
 	}
 }
 
@@ -264,7 +259,7 @@ static void decoration_manager_bind(struct wl_client *client, void *data,
 static void handle_display_destroy(struct wl_listener *listener, void *data) {
 	struct wlr_xdg_decoration_manager_v1 *manager =
 		wl_container_of(listener, manager, display_destroy);
-	wlr_signal_emit_safe(&manager->events.destroy, manager);
+	wl_signal_emit_mutable(&manager->events.destroy, manager);
 	wl_list_remove(&manager->display_destroy.link);
 	wl_global_destroy(manager->global);
 	free(manager);
